#!/usr/bin/env perl

package escapers;

sub raw
{
    return shift;
}

sub shell
{
    return &main::shell_quote(shift);
}

sub _unlang
{
    my $char = shift;
    return "\\n" if ($char eq "\n");
    return "\\r" if ($char eq "\r");
    return "\\".$char;
}

sub unlang
{
    my $string = shift;
    $string =~ s/(["\\\n\r])/_unlang($1)/ge;
    $string =~ s/\c+//g;
    return '"'.$string.'"';
}

sub json
{
    return $main::JSON->encode(shift);
}

package main;

use JSON;
use String::ShellQuote 'shell_quote';
use Getopt::Long;
use Data::Dumper;

$JSON = JSON->new;

sub createOutput
{
    my $outputFile = shift;
    if (defined $outputFile) {
        my $fh;
        if (open($fh, ">", $outputFile)) {
            return $fh;
        } else {
            die "Error: Cannot open file: $outputFile";
        }
    } else {
        return *STDOUT;
    }
}

sub createInput
{
    my $inputFile = shift;
    if (defined $inputFile) {
        my $fh;
        if (open($fh, "<", $inputFile)) {
            return $fh;
        } else {
            die "Error: Cannot open file: $inputFile";
        }
    } else {
        return *STDIN;
    }
}

sub escape
{
    my $value = shift;
    my $escaper = shift;

    return undef if not exists $escapers::{$escaper};

    return &{$escapers::{$escaper}}($value);
}

sub envsubstConst
{
    my $const = shift;
    return "{{" if $const eq 'qstart';
    return "}}" if $const eq 'qend';
    return undef;
}

sub envsubstExpr
{
    my $expr = shift;
    my $defaultEscaper = shift;
    my $value = undef;
    
    if ($expr =~ /^\s*([a-zA-Z_][a-zA-Z0-9_]*)\s*$/) {
        $value = &envsubstConst($1);
        return $value if defined $value;
    }
    if ($expr =~ /^\s*\$?([a-zA-Z_][a-zA-Z0-9_]*)\s*([:|]\s*([a-zA-Z_][a-zA-Z0-9_]*))?\s*$/) {
        my $var = $1;
        my $escaper = $3;
        $escaper = undef if $escaper eq '';

        return undef if not exists $ENV{$var};
        $value = $ENV{$var};
        $escaper = $defaultEscaper if not defined $escaper;
        $value = &escape($value, $escaper) if defined $escaper;
        return $value;
    }

    return undef;
}

sub envsubst
{
    my $tag = shift;
    my $original = shift;
    my $def = shift;
    my $defaultEscaper = shift;
    my $realTag = undef;

    if ($def =~ /^(.*)@([a-zA-Z_][a-zA-Z0-9_]*)\s*$/) {
        $realTag = $2;
        $def = $1;
    }

    return $original if (defined $tag and $tag ne $realTag);

    my $value = &envsubstExpr($def, $defaultEscaper);

    return $original if not defined $value;
    return $value;
}

sub process
{
    my $input = shift;
    my $output = shift;
    my $tag = shift;
    my $defaultEscaper = shift;
    while (<$input>) {
        s/(\{\{\s*([^\{\}]*)\}\})/envsubst($tag, $1, $2, $defaultEscaper)/ge;
        print $output $_;
    }
}

my $options = {};
GetOptions($options, "tag=s", "output=s", "escape=s");

my $tag = $options->{tag};
my $output = $options->{output};
my $defaultEscaper = $options->{escape};

my @input = ();

if (@ARGV > 0) {
    push @input, @ARGV;
} else {
    push @input, undef;
}

my $tempFile = "/tmp/envsubst-$$";
my $outputFd = &createOutput($output);

for my $input (@input) {

    my $inputFd = &createInput($input);
    my $useTemp = ((not defined $output) and (defined $input));

    if ($useTemp) {
        $outputFd = &createOutput($tempFile);
    }

    &process($inputFd, $outputFd, $tag, $defaultEscaper);

    close($inputFd);

    if ($useTemp) {
        close($outputFd);
        rename $tempFile, $input;
    }
}
